/**
 * Nehalem Uncore PAPI-C component
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, ZIH, TU Dresden, Federal Republic of Germany
 *
 * See the file COPYING in the package base directory for details
 **/

#define _GNU_SOURCE
#include <sched.h>

#include <papi.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>

/* default 0 to stress component with concurrent requests */
#define INIT_SLEEP 0
#define READ_SLEEP 1
#define NUM_THREADS 8
#define GLOBAL_EVENTSET

#define DEBUG 0

/* ARCH_AMD or ARCH_Intel */
#define ARCH_AMD

#ifdef GLOBAL_EVENTSET
int EventSet;
#ifdef ARCH_AMD
char *papi_names[4];
int papi_codes[4];
long long results[4];
#endif
#ifdef ARCH_Intel
char *papi_names[8];
int papi_codes[8];
long long results[8];
#endif
#endif

int cpu_set(int id)
{
  unsigned long long cpumask;
  cpu_set_t  mask;
  CPU_ZERO( &mask );
  CPU_SET(id , &mask );
  if (sched_setaffinity(0,sizeof(cpu_set_t),&mask))
  {
    cpumask=0|(1<<id);
    return sched_setaffinity(0,sizeof(unsigned long long),(cpu_set_t*)&cpumask);
  }
  else return 0;
}

static void *thread(void* threaddata)
{
  #ifndef GLOBAL_EVENTSET
  int EventSet;
  #ifdef ARCH_AMD
  char *papi_names[4];
  int papi_codes[4]={0,0,0,0};
  long long results[4]={0,0,0,0};
  #endif
  #ifdef ARCH_Intel
  char *papi_names[8];
  int papi_codes[8]={0,0,0,0,0,0,0,0};
  long long results[8]={0,0,0,0,0,0,0,0};
  #endif
  #endif
  int i;

  cpu_set(*((int*)threaddata));
  
  #ifndef GLOBAL_EVENTSET
  if (DEBUG) {printf("hello from Thread %i\n",*((int*)threaddata));fflush(stdout);}
  
  #ifdef ARCH_AMD
  for (i=0;i<4;i++) papi_names[i]=malloc(256);
  sprintf(papi_names[0],"SYSWIDE_L3_CACHE_MISSES:READ_BLOCK_EXCLUSIVE:CORE_0_SELECT:CORE_1_SELECT");
  sprintf(papi_names[1],"SYSWIDE_READ_REQUEST_TO_L3_CACHE:ALL");
  #endif
  #ifdef ARCH_Intel
  for (i=0;i<8;i++) papi_names[i]=malloc(256);
  sprintf(papi_names[0],"SYSWIDE_UNC_GQ_CYCLES_NOT_EMPTY:READ_TRACKER");
  sprintf(papi_names[1],"SYSWIDE_UNC_GQ_CYCLES_FULL:WRITE_TRACKER");
  sprintf(papi_names[2],"SYSWIDE_UNC_GQ_CYCLES_NOT_EMPTY:WRITE_TRACKER");
  #endif
  
  EventSet = PAPI_NULL;
  if (PAPI_create_eventset(&EventSet) != PAPI_OK)
  {
      printf("Warning: PAPI error, could not create eventset\n");fflush(stdout);
      pthread_exit(0);
  }
  if (DEBUG) {printf("Thread %i created eventset %i\n",*((int*)threaddata),EventSet);fflush(stdout);}
 
  #ifdef ARCH_AMD
  for (i=0;i<2;i++)
  #endif
  #ifdef ARCH_Intel
  for (i=0;i<3;i++)
  #endif
  {
    //printf("add event %i\n",i);
    if (PAPI_event_name_to_code(papi_names[i],&papi_codes[i])!=PAPI_OK)
    {
       printf("Warning: unknown Counter: %s\n",papi_names[i]);fflush(stdout);
       pthread_exit(0);
    }
    else
    { 
      if ((PAPI_add_event(EventSet, papi_codes[i]) != PAPI_OK))
      {
        printf("Warning: PAPI error, could not add counter %s to eventset.\n",papi_names[i]);
        pthread_exit(0);
      }
      else if (DEBUG) printf("added event %s(%8x)\n",papi_names[i],papi_codes[i]);
    }
  }
  if (DEBUG) {printf("Thread %i added events, starting measurement\n",*((int*)threaddata));fflush(stdout);}
  if (PAPI_start(EventSet)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_start error"); 
  #else 
  if (DEBUG) {printf("Thread %i starting measurement\n",*((int*)threaddata));fflush(stdout);}
  if (DEBUG) {printf("Thread %i starting measurement\n",PAPI_thread_id());fflush(stdout);}
  #endif

  sleep(READ_SLEEP);
  if (DEBUG) printf("Thread %i - PAPI_read\n",*((int*)threaddata));
  if (PAPI_read(EventSet,results)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_read error");

  #ifdef ARCH_AMD
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx\n",*((int*)threaddata),results[0],results[1],results[2],results[3]);
  #endif
  #ifdef ARCH_Intel
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx¸n",*((int*)threaddata),results[0],results[1],results[2],results[3],results[4],results[5],results[6],results[7]);
  #endif
  sleep(READ_SLEEP);
  cpu_set(((*((int*)threaddata))+1)%8);
  if (DEBUG) printf("Thread %i - PAPI_read\n",*((int*)threaddata));
  if (PAPI_read(EventSet,results)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_read error"); 

  #ifdef ARCH_AMD
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx\n",*((int*)threaddata),results[0],results[1],results[2],results[3]);
  #endif
  #ifdef ARCH_Intel
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx\n",*((int*)threaddata),results[0],results[1],results[2],results[3],results[4],results[5],results[6],results[7]);
  #endif
  sleep(READ_SLEEP);
  cpu_set(((*((int*)threaddata))+4)%8);
  if (DEBUG) printf("Thread %i - PAPI_read\n",*((int*)threaddata));
  if (PAPI_read(EventSet,results)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_read error"); 

  #ifdef ARCH_AMD
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx\n",*((int*)threaddata),results[0],results[1],results[2],results[3]);
  #endif
  #ifdef ARCH_Intel
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx\n",*((int*)threaddata),results[0],results[1],results[2],results[3],results[4],results[5],results[6],results[7]);
  #endif
  sleep(READ_SLEEP);
  #ifdef GLOBAL_EVENTSET
  if (DEBUG) printf("Thread %i - PAPI_read\n",*((int*)threaddata));
  if (PAPI_read(EventSet,results)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_read error");
  #else
  if (DEBUG) printf("Thread %i - PAPI_stop\n",*((int*)threaddata));  
  if (PAPI_stop(EventSet,results)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_stop error"); 
  #endif
  #ifdef ARCH_AMD
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx\n",*((int*)threaddata),results[0],results[1],results[2],results[3]);
  #endif
  #ifdef ARCH_Intel
  printf("Thread %i: %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx, %016llx\n",*((int*)threaddata),results[0],results[1],results[2],results[3],results[4],results[5],results[6],results[7]);
  #endif
  if (DEBUG) {printf("Thread %i stoped measurement and exits\n",*((int*)threaddata));fflush(stdout);}

  pthread_exit(0);
}

int main(int argc, char **argv)
{
 pthread_t threads[8];
 int ids[8]={0,1,2,3,4,5,6,7};
 int i;

 if (PAPI_library_init(PAPI_VER_CURRENT) != PAPI_VER_CURRENT) printf("Warning: PAPI_library_init error\n");

 else if (DEBUG) printf("PAPI_library_init OK\n"); 

 if (PAPI_thread_init(pthread_self)!=PAPI_OK) { if (DEBUG)printf("Warning: PAPI_thread_init error\n");return -1;}
 else if (DEBUG) printf("PAPI_thread_init OK\n");

 #ifdef GLOBAL_EVENTSET

  #ifdef ARCH_AMD
  for (i=0;i<4;i++) papi_names[i]=malloc(256);
  sprintf(papi_names[0],"SYSWIDE_L3_CACHE_MISSES:READ_BLOCK_EXCLUSIVE:CORE_0_SELECT:CORE_1_SELECT");
  sprintf(papi_names[1],"SYSWIDE_READ_REQUEST_TO_L3_CACHE:ALL");
  #endif
  #ifdef ARCH_Intel
  for (i=0;i<8;i++) papi_names[i]=malloc(256);
  sprintf(papi_names[0],"SYSWIDE_UNC_GQ_CYCLES_NOT_EMPTY:READ_TRACKER");
  sprintf(papi_names[1],"SYSWIDE_UNC_GQ_CYCLES_FULL:WRITE_TRACKER");
  sprintf(papi_names[2],"SYSWIDE_UNC_GQ_CYCLES_NOT_EMPTY:WRITE_TRACKER");
  #endif

  EventSet = PAPI_NULL;
  if (PAPI_create_eventset(&EventSet) != PAPI_OK)
  {
      printf("Warning: PAPI error, could not create eventset\n");fflush(stdout);
      pthread_exit(0);
  }
  if (DEBUG) {printf("globally created eventset %i\n",EventSet);fflush(stdout);}

  for (i=0;i<2;i++)
  {
    //printf("add event %i\n",i);
    if (PAPI_event_name_to_code(papi_names[i],&papi_codes[i])!=PAPI_OK)
    {
       printf("Warning: unknown Counter: %s\n",papi_names[i]);fflush(stdout);
       pthread_exit(0);
    }
    else
    { 
      if ((PAPI_add_event(EventSet, papi_codes[i]) != PAPI_OK))
      {
        printf("Warning: PAPI error, could not add counter %s to eventset.\n",papi_names[i]);
        pthread_exit(0);
      }
      else if (DEBUG) printf("added event %s(%8x)\n",papi_names[i],papi_codes[i]);
    }
  }
  if (DEBUG) {printf("added events\n");fflush(stdout);}
 if (PAPI_start(EventSet)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_start error");
 #endif

 for (i=0;i<NUM_THREADS;i++) {sleep(INIT_SLEEP);pthread_create(&threads[i],NULL,thread,(void*) &ids[i]);}
 
 for (i=0;i<NUM_THREADS;i++) pthread_join(threads[i],NULL);

 #ifdef GLOBAL_EVENTSET
 if (PAPI_stop(EventSet,results)!=PAPI_OK) if (DEBUG) perror("Warning: PAPI_stop error"); 
 #endif

 return 0;
}
